// fair4ctf.cpp : Defines the entry point for the DLL application.
//

#include "bzfsAPI.h"
#include <algorithm>
#include <cmath>

BZ_GET_PLUGIN_VERSION

class FCTF {
public:
	FCTF() {
		fairCTFEnabled = true; // turn on/off this plugin
		notifiedCTFOK_RG = false;
		notifiedCTFOK_RB = false;
		notifiedCTFOK_RP = false;
		notifiedCTFOK_GB = false;
		notifiedCTFOK_GP = false;
		notifiedCTFOK_BP = false;
		fairCTF = true;  // all teams are fair = true
		kickEnabled = false; // auto-kick is off by default, uses /kill instead

		RvG = 0;
		RvB = 0;
		RvP = 0;
		GvB = 0;	
		GvP = 0;
		BvP = 0;

		// don't worry about teams with max 0 players
		red = (bz_getTeamPlayerLimit((bz_eTeamType)1) != 0);
		green = (bz_getTeamPlayerLimit((bz_eTeamType)2) != 0);
		blue = (bz_getTeamPlayerLimit((bz_eTeamType)3) != 0);
		purple = (bz_getTeamPlayerLimit((bz_eTeamType)4) != 0);
		
		// kick player after this may illegal grabs
		illegalGrabThreshold = 3;

		obsMsg = false;
	}

	int RvG;	//  1: left is strong (ie, red > green)
	int RvB;	//  0: teams are even (ie, red == blue)
	int RvP;	// -1: right is strong (ie, purple > red)
	int GvB;	
	int GvP;
	int BvP;
	bool red;
	bool green;
	bool blue;
	bool purple;
	std::vector<std::pair<int,int> > illegalGrabs;
	int illegalGrabThreshold;

	bool fairCTFEnabled;
	bool notifiedCTFOK_RG;
	bool notifiedCTFOK_RB;
	bool notifiedCTFOK_RP;
	bool notifiedCTFOK_GB;
	bool notifiedCTFOK_GP;
	bool notifiedCTFOK_BP;
	bool fairCTF;
	bool kickEnabled;
	bool obsMsg;
};

FCTF fctf;

class FCTFPlayerUpdates : public bz_EventHandler {
public:
	virtual void	process ( bz_EventData *eventData );
};

class FCTFTickEvents : public bz_EventHandler {
public:
	virtual void	process ( bz_EventData *eventData );
};

class FCTFIllegalGrabKick : public bz_EventHandler {
public:
	virtual void	process ( bz_EventData *eventData );
};

class FCTFPlayerPart : public bz_EventHandler {
public:
	virtual void	process ( bz_EventData *eventData );
};

class FCTFCommands : public bz_CustomSlashCommandHandler {
public:
  virtual ~FCTFCommands(){};
  virtual bool handle ( int playerID, bzApiString command, bzApiString message, bzAPIStringList *param );
};

class FCTFPlayerJoined : public bz_EventHandler {
public:
	virtual void	process ( bz_EventData *eventData );
};

FCTFPlayerUpdates fctfplayerupdates;
FCTFIllegalGrabKick fctfillegalgrabkick;
FCTFPlayerPart fctfplayerpart;
FCTFTickEvents fctftickevents;
FCTFCommands fctfcommands;
FCTFPlayerJoined fctfplayerjoined;


// use the input to set illegalGrabThreshold
int ConvertToInt(std::string inmessage) {
	int messagelength = (int)inmessage.length();
	if (messagelength > 0 && messagelength < 3){
		int messagevalue = 0;
		int tens = 1;
		for ( int i = (messagelength - 1); i >= 0; i-- ){
			if (inmessage[i] < '0' || inmessage[i] > '9')  // got something other than a number
				return -1; 
			tens *= 10;
			messagevalue +=  (int) ((((double)inmessage[i] - '0') / 10) * tens);
		}
		// we're ok so far, check limits:
		if (messagevalue >= 0 && messagevalue <= 99)
			return messagevalue;
	}
	return -1;
}

BZF_PLUGIN_CALL int bz_Load ( const char* commandLine ) {
  bz_debugMessage(4,"Fair4CTF plugin loaded");

  std::string parameter = commandLine;
  int kickThreshold = ConvertToInt(parameter);
  
  if (kickThreshold != -1) //if it is -1, leave set to the default
	  fctf.illegalGrabThreshold = kickThreshold;

  bz_registerEvent(bz_ePlayerJoinEvent, &fctfplayerjoined);
  bz_registerEvent(bz_ePlayerUpdateEvent, &fctfplayerupdates);
  if (fctf.illegalGrabThreshold != 0) {
	  bz_registerEvent(bz_eFlagGrabbedEvent, &fctfillegalgrabkick); //new
	  bz_registerEvent(bz_ePlayerPartEvent, &fctfplayerpart); //new
  }
  bz_registerEvent(bz_eTickEvent,&fctftickevents);

  //FIXME (with line 156)
  bz_registerCustomSlashCommand("ctf",&fctfcommands);
  return 0;
}

BZF_PLUGIN_CALL int bz_Unload ( void ) {
  bz_debugMessage(4,"Fair4CTF plugin unloaded");
  bz_removeEvent(bz_ePlayerJoinEvent, &fctfplayerjoined);
  bz_removeEvent(bz_ePlayerUpdateEvent, &fctfplayerupdates);
  if (fctf.illegalGrabThreshold != 0) {
	bz_removeEvent(bz_eFlagGrabbedEvent,&fctfillegalgrabkick); //new
	bz_removeEvent(bz_ePlayerPartEvent,&fctfplayerpart); //new
  }

  //FIXME (with line 142)
  bz_removeEvent(bz_eTickEvent,&fctftickevents);
  bz_removeCustomSlashCommand("ctf");
  return 0;
}


// check for fairness between each pair of teams:
bool TeamsBalanced() {

	float TeamRatioTolerance = 0.25; // if not same size, at least 3 versus 4 or better
	
	// compare teams in pairs, ratio is (team difference)/(larger team)
	float RatioRG = 0;
	float RatioRB = 0;
	float RatioRP = 0;
	float RatioGB = 0;
	float RatioBP = 0;
	float RatioGP = 0;
	// get current team sizes
	float RS = (float)bz_getTeamCount(eRedTeam);
	float GS = (float)bz_getTeamCount(eGreenTeam);
	float BS = (float)bz_getTeamCount(eBlueTeam);
	float PS = (float)bz_getTeamCount(ePurpleTeam);

	// calculate ratios
	if (std::max(RS, GS) != 0 && fctf.red && fctf.green)
		RatioRG = (RS-GS)/std::max(RS, GS);
	if (std::max(RS, BS) != 0 && fctf.red && fctf.blue)
		RatioRB = (RS-BS)/std::max(RS, BS);
	if (std::max(RS, PS) != 0 && fctf.red && fctf.purple)
		RatioRP = (RS-PS)/std::max(RS, PS);
	if (std::max(GS, BS) != 0 && fctf.green && fctf.blue)
		RatioGB = (GS-BS)/std::max(GS, BS);
	if (std::max(GS, PS) != 0 && fctf.green && fctf.purple)
		RatioGP = (GS-PS)/std::max(GS, PS);
	if (std::max(BS, PS) != 0 && fctf.blue && fctf.purple)
		RatioBP = (BS-PS)/std::max(BS, PS);
	
	// use ratios and tolerance to determine if the pair is fair
	// if (AvB == 1) A is stronger than B
	// if (AvB == 0) A and B are close enough to be even
	// if (AvB == -1) B is stronger than A
	if (fabs(RatioRG) <= TeamRatioTolerance)
		fctf.RvG = 0;
	else {
		if (RatioRG > 0)
			fctf.RvG = 1;
		else
			fctf.RvG = -1;
	}
	if (fabs(RatioRB) <= TeamRatioTolerance)
		fctf.RvB = 0;
	else {
		if (RatioRB > 0)
			fctf.RvB = 1;
		else
			fctf.RvB = -1;
	}
	if (fabs(RatioRP) <= TeamRatioTolerance)
		fctf.RvP = 0;
	else {
		if (RatioRP > 0)
			fctf.RvP = 1;
		else
			fctf.RvP = -1;
	}
	if (fabs(RatioGB) <= TeamRatioTolerance)
		fctf.GvB = 0;
	else {
		if (RatioGB > 0)
			fctf.GvB = 1;
		else
			fctf.GvB = -1;
	}
	if (fabs(RatioGP) <= TeamRatioTolerance)
		fctf.GvP = 0;
	else {
		if (RatioGP > 0)
			fctf.GvP = 1;
		else
			fctf.GvP = -1;
	}
	if (fabs(RatioBP) <= TeamRatioTolerance)
		fctf.BvP = 0;
	else {
		if (RatioBP > 0)
			fctf.BvP = 1;
		else
			fctf.BvP = -1;
	}

	// returns ctf as fair ONLY if every pair is fair
	if ( (abs(fctf.RvG) + abs(fctf.RvB) + abs(fctf.RvP) + abs(fctf.GvB) + abs(fctf.GvP) + abs(fctf.BvP)) == 0)
		return true;
	else
		return false;
}


// notifies joining players the current ctf status (allowed/disallowed)
void FCTFPlayerJoined::process ( bz_EventData *eventData ) {
	if (eventData->eventType != bz_ePlayerJoinEvent || !fctf.fairCTFEnabled)
		return;

	bz_PlayerJoinPartEventData *JoinData = (bz_PlayerJoinPartEventData*)eventData;

	// if teams are not even, notify joiner which teams CTF is currently disabled for
	switch(JoinData->team) {
		case eRedTeam:	
			if ( fctf.RvG<=0 && fctf.RvB<=0 && fctf.RvP <= 0) {	// fair teams
				bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "CTF enabled");
			} else {
				bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "CTF disabled against:");
				if (fctf.RvG == 1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Green");
				if (fctf.RvB == 1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Blue");
				if (fctf.RvP == 1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Purple");
			}
			break;
		case eGreenTeam:
			if ( fctf.RvG<=0 && fctf.GvB<=0 && fctf.GvP <= 0) {	// fair teams
				bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "CTF enabled");
			} else {
				bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "CTF disabled against:");
				if (fctf.RvG == -1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Red");
				if (fctf.GvB == 1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Blue");
				if (fctf.GvP == 1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Purple");
			}
			break;
		case eBlueTeam:
			if ( fctf.RvB<=0 && fctf.GvB<=0 && fctf.BvP <= 0) {	// fair teams
				bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "CTF enabled");
			} else {
				bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "CTF disabled against:");
				if (fctf.RvB == -1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Red");
				if (fctf.GvB == -1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Green");
				if (fctf.BvP == 1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Purple");
			}
			break;
		case ePurpleTeam:
			if ( fctf.RvP<=0 && fctf.GvP<=0 && fctf.BvP <= 0) {	// fair teams
				bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "CTF enabled");
			} else {
				bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "CTF disabled against:");
				if (fctf.RvP == -1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Red");
				if (fctf.GvP == -1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Green");
				if (fctf.BvP == -1) bz_sendTextMessagef (BZ_SERVER, JoinData->playerID, "Blue");
			}
			break;
	}
	return;
}


// notify both teams & observers when CTF status changes
void FCTFTickEvents::process ( bz_EventData *eventData ) {
	if (eventData->eventType != bz_eTickEvent || !fctf.fairCTFEnabled)
		return;

	std::string obsOn, obsOff;
	obsOn = "";
	obsOff = "";

	// read this function once per tick event to update fctf.?v?  
	fctf.fairCTF = TeamsBalanced();  

	// check/notify team balance changes
	// if fair ctf is disabled, no need to check/notify about team balance changes:

	// messages to allow CTF
	if (fctf.RvG == 0 && !fctf.notifiedCTFOK_RG  && fctf.red && fctf.green) { // red vs green is OK
		bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Capture the flag enabled for Red vs Green - teams are fair enough.");
		bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Capture the flag enabled for Green vs Red - teams are fair enough.");
		obsOn += "Red vs Green, ";
		fctf.notifiedCTFOK_RG = true;
	}
	if (fctf.RvB == 0 && !fctf.notifiedCTFOK_RB  && fctf.red && fctf.blue) { // red vs blue is OK
		bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Capture the flag enabled for Red vs Blue - teams are fair enough.");
		bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Capture the flag enabled for Blue vs Red - teams are fair enough.");
		obsOn += "Red vs Blue, ";
		fctf.notifiedCTFOK_RB = true;
	}
	if (fctf.RvP == 0 && !fctf.notifiedCTFOK_RP  && fctf.red && fctf.purple) { // red vs purple is OK
		bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Capture the flag enabled for Red vs Purple - teams are fair enough.");
		bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Capture the flag enabled for Purple vs Red - teams are fair enough.");
		obsOn += "Red vs Purple, ";
		fctf.notifiedCTFOK_RP = true;
	}
	if (fctf.GvB == 0 && !fctf.notifiedCTFOK_GB  && fctf.green && fctf.blue) { // green vs blue is OK
		bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Capture the flag enabled for Green vs Blue - teams are fair enough.");
		bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Capture the flag enabled for Blue vs Green - teams are fair enough.");
		obsOn += "Green vs Blue, ";
		fctf.notifiedCTFOK_GB = true;
	}
	if (fctf.GvP == 0 && !fctf.notifiedCTFOK_GP  && fctf.green && fctf.purple) { // green vs purple is OK
		bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Capture the flag enabled for Green vs Purple - teams are fair enough.");
		bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Capture the flag enabled for Purple vs Green - teams are fair enough.");
		obsOn += "Green vs Purple, ";
		fctf.notifiedCTFOK_GP = true;
	}
	if (fctf.BvP == 0 && !fctf.notifiedCTFOK_BP  && fctf.blue && fctf.purple) { // blue vs purple is OK
		bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Capture the flag enabled for Blue vs Purple - teams are fair enough.");
		bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Capture the flag enabled for Purple vs Blue - teams are fair enough.");
		obsOn += "Blue vs Purple, ";
		fctf.notifiedCTFOK_BP = true;
	}
	
	// messages to disable CTF
	if (fctf.notifiedCTFOK_RG  && fctf.red && fctf.green) { // red vs green is uneven
		if (fctf.RvG > 0){
			bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Capture the flag disabled against Green - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Green flag is safe from Red - teams are uneven.");
			obsOff += "Red vs Green, ";
			fctf.notifiedCTFOK_RG = false;
		}
		if (fctf.RvG < 0){
			bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Capture the flag disabled against Red - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Red flag is safe from Green - teams are uneven.");
			obsOff += "Green vs Red, ";
			fctf.notifiedCTFOK_RG = false;
		}
	}
	if (fctf.notifiedCTFOK_RB  && fctf.red && fctf.blue) { // red vs blue is uneven
		if (fctf.RvB > 0){
			bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Capture the flag disabled against Blue - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Blue flag is safe from Red - teams are uneven.");
			obsOff += "Red vs Blue, ";
			fctf.notifiedCTFOK_RB = false;
		}
		if (fctf.RvB < 0){
			bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Capture the flag disabled against Red - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Red flag is safe from Blue - teams are uneven.");
			obsOff = obsOff+ "Blue vs Red, ";
			fctf.notifiedCTFOK_RB = false;
		}
	}
	if (fctf.notifiedCTFOK_RP  && fctf.red && fctf.purple) { // red vs purple is uneven
		if (fctf.RvP > 0){
			bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Capture the flag disabled against Purple - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Purple flag is safe from Red - teams are uneven.");
			obsOff += "Red vs Purple, ";
			fctf.notifiedCTFOK_RP = false;
		}
		if (fctf.RvP < 0){
			bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Capture the flag disabled against Red - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eRedTeam, "Red flag is safe from Purple - teams are uneven.");
			obsOff += "Purple vs Red, ";
			fctf.notifiedCTFOK_RP = false;
		}
	}
	if (fctf.notifiedCTFOK_GB  && fctf.green && fctf.blue) { // green vs blue is uneven
		if (fctf.GvB > 0){
			bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Capture the flag disabled against Blue - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Blue flag is safe from Green - teams are uneven.");
			obsOff += "Green vs Blue, ";
			fctf.notifiedCTFOK_GB = false;
		}
		if (fctf.GvB < 0){
			bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Capture the flag disabled against Green - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Green flag is safe from Blue - teams are uneven.");
			obsOff += "Blue vs Green, ";
			fctf.notifiedCTFOK_GB = false;
		}
	}
	if (fctf.notifiedCTFOK_GP  && fctf.green && fctf.purple) { // green vs purple is uneven
		if (fctf.GvP > 0){
			bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Capture the flag disabled against Purple - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Purple flag is safe from Green - teams are uneven.");
			obsOff += "Green vs Purple, ";
			fctf.notifiedCTFOK_GP = false;
		}
		if (fctf.GvP < 0){
			bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Capture the flag disabled against Green - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eGreenTeam, "Green flag is safe from Purple - teams are uneven.");
			obsOff += "Purple vs Green, ";
			fctf.notifiedCTFOK_GP = false;
		}
	}
	if (fctf.notifiedCTFOK_BP  && fctf.blue && fctf.purple) { // blue vs purple is uneven
		if (fctf.BvP > 0){
			bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Capture the flag disabled against Purple - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Purple flag is safe from Blue - teams are uneven.");
			obsOff += "Blue vs Purple, ";
			fctf.notifiedCTFOK_BP = false;
		}
		if (fctf.BvP < 0){
			bz_sendTextMessagef (BZ_SERVER, ePurpleTeam, "Capture the flag disabled against Blue - teams are uneven.");
			bz_sendTextMessagef (BZ_SERVER, eBlueTeam, "Blue flag is safe from Purple - teams are uneven.");
			obsOff += "Purple vs Blue, ";
			fctf.notifiedCTFOK_BP = false;
		}
	}
	if (fctf.obsMsg) {
		std::string output = "";
		if (obsOn.length() != 0) {
			obsOn = obsOn.substr(0, obsOn.length() - 2);
			output += "CTF enabled for: " + obsOn + ".   ";
		}
		if (obsOff.length() != 0) {
			obsOff = obsOff.substr(0, obsOff.length() - 2);
			output += "CTF disabled for: " + obsOff + ".";
		}
		if (output.length() != 0)
			bz_sendTextMessagef(BZ_SERVER, eObservers, output.c_str());
	}

	return;
}


//  check and force drop team flags when ctf is uneven
void FCTFPlayerUpdates::process ( bz_EventData *eventData ) {
	if (eventData->eventType != bz_ePlayerUpdateEvent || !fctf.fairCTFEnabled)
		return;

	// no CTF if teams not balanced, drop team flags for the stronger team(s):

	if (!fctf.fairCTF){
		const int playerID = ((bz_PlayerUpdateEventData*)eventData)->playerID;
		bz_PlayerRecord *record;
		record = bz_getPlayerByIndex(playerID);
		const int team = record->team;
		const char* FlagHeld = bz_getPlayerFlag(playerID);
		bool warnPlayer = false;

		if (FlagHeld != NULL) {
			switch (team) {
				case eRedTeam:
					if (strcmp(FlagHeld, "G*") == 0 && (fctf.RvG == 1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Green - teams are uneven.");
						warnPlayer=true;
					}
					if (strcmp(FlagHeld, "B*") == 0 && (fctf.RvB == 1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Blue - teams are uneven.");
						warnPlayer=true;
					}
					if (strcmp(FlagHeld, "P*") == 0 && (fctf.RvP == 1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Purple - teams are uneven.");
						warnPlayer=true;
					}
					break;
				case eGreenTeam:
					if (strcmp(FlagHeld, "R*") == 0 && (fctf.RvG == -1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Red - teams are uneven.");
						warnPlayer=true;
					}
					if (strcmp(FlagHeld, "B*") == 0 && (fctf.GvB == 1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Blue - teams are uneven.");
						warnPlayer=true;
					}
					if (strcmp(FlagHeld, "P*") == 0 && (fctf.GvP == 1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Purple - teams are uneven.");
						warnPlayer=true;
					}
					break;
				case eBlueTeam:
					if (strcmp(FlagHeld, "R*") == 0 && (fctf.RvB == -1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Red - teams are uneven.");
						warnPlayer=true;
					}
					if (strcmp(FlagHeld, "G*") == 0 && (fctf.GvB == -1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Green - teams are uneven.");
						warnPlayer=true;
					}
					if (strcmp(FlagHeld, "P*") == 0 && (fctf.BvP == 1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Purple - teams are uneven.");
						warnPlayer=true;
					}
					break;
				case ePurpleTeam:
					if (strcmp(FlagHeld, "R*") == 0 && (fctf.RvP == -1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Red - teams are uneven.");
						warnPlayer=true;
					}
					if (strcmp(FlagHeld, "G*") == 0 && (fctf.GvP == -1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Green - teams are uneven.");
						warnPlayer=true;
					}
					if (strcmp(FlagHeld, "B*") == 0 && (fctf.BvP == -1)) {
						bz_removePlayerFlag ( playerID );
						bz_sendTextMessagef (BZ_SERVER, playerID, "Capture the flag is disabled against Blue - teams are uneven.");
						warnPlayer=true;
					}
					break;
			}
			if (warnPlayer && fctf.illegalGrabThreshold != 0) {
				if (fctf.kickEnabled) 
					bz_sendTextMessagef(BZ_SERVER, playerID, "Repeatedly grabbing a weaker team's flag WILL get you kicked.");
				else
					bz_sendTextMessagef(BZ_SERVER, playerID, "Repeatedly grabbing a weaker team's flag WILL get you /killed");
			}
			bz_freePlayerRecord (record);
		}
	}
	return;
}


// keep track of how many uneven flag pickups each player has, and kick those with too many.
void FCTFIllegalGrabKick::process ( bz_EventData *eventData ) {	
	bool flagGrab = (eventData->eventType == bz_eFlagGrabbedEvent);
	bool logForKick = false;
	if (!flagGrab || !fctf.fairCTFEnabled) 
		return;

	const int playerID = ((bz_FlagGrabbedEventData*)eventData)->playerID;
	bz_PlayerRecord *record;
	record = bz_getPlayerByIndex(playerID);
	const int team = record->team;
	const char* FlagHeld = bz_getPlayerFlag(playerID);

	int teamFlag = 0;
	if (strcmp(FlagHeld, "R*") == 0) teamFlag = eRedTeam;
	if (strcmp(FlagHeld, "G*") == 0) teamFlag = eGreenTeam;
	if (strcmp(FlagHeld, "B*") == 0) teamFlag = eBlueTeam;
	if (strcmp(FlagHeld, "P*") == 0) teamFlag = ePurpleTeam;


	if (FlagHeld != NULL) {
		// set logForKick to true if the flag grabbed belongs to a weaker team
		switch (team) {
			case eRedTeam:
				logForKick = ((fctf.RvG == 1) && (teamFlag==eGreenTeam)) ||
							 ((fctf.RvB == 1) && (teamFlag==eBlueTeam)) ||
							 ((fctf.RvP == 1) && (teamFlag==ePurpleTeam));
				break;
			case eGreenTeam:
				logForKick = ((fctf.RvG == -1) && (teamFlag==eRedTeam)) ||
							 ((fctf.GvB == 1) && (teamFlag==eBlueTeam)) ||
							 ((fctf.GvP == 1) && (teamFlag==ePurpleTeam));
				break;
			case eBlueTeam:
				logForKick = ((fctf.RvB == -1) && (teamFlag==eRedTeam)) ||
							 ((fctf.GvB == -1) && (teamFlag==eGreenTeam)) ||
							 ((fctf.BvP == 1) && (teamFlag==ePurpleTeam));
				break;
			case ePurpleTeam:
				logForKick = ((fctf.RvP == -1) && (teamFlag==eRedTeam)) ||
							 ((fctf.GvP == -1) && (teamFlag==eGreenTeam)) ||
							 ((fctf.BvP == -1) && (teamFlag==eBlueTeam));
				break;
		}
	}

	if (logForKick && fctf.illegalGrabThreshold!=0) {
		// check list for playerID
			// if it's there, add 1 to the count, and check if it's over the limit to kick
			// otherwise, add the playerID to the list and set count to 1
		std::vector< std::pair<int,int> >::iterator iter;
		bool found = false;
		for( iter = fctf.illegalGrabs.begin(); iter != fctf.illegalGrabs.end(); iter++ ) {
			if ((*iter).first == playerID) {
				found = true;
				(*iter).second++;
				if (iter < fctf.illegalGrabs.end() && (*iter).second >= fctf.illegalGrabThreshold){
					fctf.illegalGrabs.erase(iter);  // player is gone, remove them from list
					if (fctf.kickEnabled)
						bz_kickUser(playerID, "Repeatedly grabbing a weaker team's flag when CTF is unfair", true);
					else
						bz_killPlayer(playerID, true, BZ_SERVER);
				}
				break;
			}
		}
		if (!found) {
			iter = fctf.illegalGrabs.end();
			std::pair<int,int> data;
			data.first = playerID;
			data.second = 1;
			fctf.illegalGrabs.insert(iter, data);
		}

	} else { //if it's a valid flag grab, clear the player's illegal grab count
		if (teamFlag != 0 ) {
			std::vector< std::pair<int,int> >::iterator iter;
			for( iter = fctf.illegalGrabs.begin(); iter != fctf.illegalGrabs.end(); iter++ ) {
				if ((*iter).first == playerID && iter < fctf.illegalGrabs.end() ) {
					fctf.illegalGrabs.erase(iter);  // remove player from list
					break;
				}
			}
		}
	}
	bz_freePlayerRecord (record);

}


// when a player leaves, remove them from the illegal grabs list
void FCTFPlayerPart::process ( bz_EventData *eventData ) {	
	if (eventData->eventType != bz_ePlayerPartEvent)
		return;
	const int playerID = (( bz_PlayerJoinPartEventData*)eventData)->playerID;
	std::vector< std::pair<int,int> >::iterator iter;
	for( iter = fctf.illegalGrabs.begin(); iter != fctf.illegalGrabs.end(); iter++ ) {
		if ((*iter).first == playerID && iter < fctf.illegalGrabs.end() ) {
			fctf.illegalGrabs.erase(iter);  //remove player from list
			break;
		}
	}
}

bool FCTFCommands::handle ( int playerID, bzApiString _command, bzApiString _message, bzAPIStringList * /*_param*/ ) {

	std::string command = _command.c_str();
	std::string message = _message.c_str();

	bz_PlayerRecord *record;
	record = bz_getPlayerByIndex(playerID);
	const int team = record->team;


	if (!bz_hasPerm(playerID, "FAIRCTF")) {
		bz_sendTextMessagef(BZ_SERVER, playerID, "You do not have permission to run the /ctf command.");
	} else {
		if (message == "on") {
			if (!fctf.fairCTFEnabled) {
				bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "Fair4CTF plugin enabled.");
				fctf.fairCTFEnabled = true;
			} else 
				bz_sendTextMessagef(BZ_SERVER, playerID, "Fair4CTF plugin is already running.");
		} else if (message == "off") {
			if (fctf.fairCTFEnabled) {
				bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "Fair4CTF plugin disabled.");
				fctf.fairCTFEnabled = false;
				// flush the stored player list of illegal grabs
				fctf.illegalGrabs.clear();
			} else 
				bz_sendTextMessagef(BZ_SERVER, playerID, "Fair4CTF plugin is already stopped.");
		} else if (message == "kick") {
			if (!fctf.kickEnabled) {
				bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "Fair4CTF set to /kick on illegal flag grabs.");
				fctf.kickEnabled = true;
			} else 
				bz_sendTextMessagef(BZ_SERVER, playerID, "Fair4CTF is already set to /kick.");
		} else	if (message == "kill") {
			if (fctf.kickEnabled) {
				bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "Fair4CTF set to /kill on illegal flag grabs.");
				fctf.kickEnabled = false;
			} else 
				bz_sendTextMessagef(BZ_SERVER, playerID, "Fair4CTF is already set to /kill.");
		} else if (message == "msgon") {
			if (!fctf.obsMsg) {
				bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "Fair4CTF observer messages are enabled.");
				fctf.obsMsg = true;
			} else 
				bz_sendTextMessagef(BZ_SERVER, playerID, "Fair4CTF observer messages are already enabled.");
		} else if (message == "msgoff") {
			if (fctf.obsMsg) {
				bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "Fair4CTF observer messages are disabled.");
				fctf.obsMsg = false;
			} else 
				bz_sendTextMessagef(BZ_SERVER, playerID, "Fair4CTF observer messages are already disabled.");
		} else if (message.length() >= 5 && message.substr(0,5) == "grabs") {
			if (message.length() <= 6)
				bz_sendTextMessagef(BZ_SERVER, playerID, "Illegal grab limit: %d", fctf.illegalGrabThreshold);
			else {
				int param = ConvertToInt(message.substr(6, message.length()-6));
				if (param == -1) 
					bz_sendTextMessagef(BZ_SERVER, playerID, "Illegal grab limit: %d", fctf.illegalGrabThreshold);
				else {
					fctf.illegalGrabThreshold = param;
					bz_sendTextMessagef(BZ_SERVER, BZ_ALLUSERS, "Illegal grab limit set to %d", fctf.illegalGrabThreshold);
				}
			}
		} else
			bz_sendTextMessagef(BZ_SERVER, playerID, "Usage: /ctf (on|off|kick|kill|msgon|msgoff|grabs [#])");
		}
	bz_freePlayerRecord (record);
	return true;
}



// Local Variables: ***
// mode:C++ ***
// tab-width: 8 ***
// c-basic-offset: 2 ***
// indent-tabs-mode: t ***
// End: ***
// ex: shiftwidth=2 tabstop=8

